from __future__ import annotations

from alphafold3_pytorch.tensor_typing import typecheck
from typing import Callable, List, Dict, Literal

from alphafold3_pytorch.alphafold3 import Alphafold3

from alphafold3_pytorch.inputs import (
    AtomDataset,
    PDBDataset,
    pdb_dataset_to_atom_inputs
)

from alphafold3_pytorch.trainer import (
    Trainer,
    Dataset,
    Fabric,
    Optimizer,
    LRScheduler
)

from alphafold3_pytorch.data.weighted_pdb_sampler import WeightedPDBSampler

import yaml
from pathlib import Path

from pydantic import BaseModel, model_validator

from pydantic.types import (
    FilePath,
    DirectoryPath
)

# functions

def exists(v):
    return v is not None

@typecheck
def safe_deep_get(
    d: dict,
    dotpath: str | List[str],  # dotpath notation, so accessing {'a': {'b'': {'c': 1}}} would be "a.b.c"
    default = None
):
    if isinstance(dotpath, str):
        dotpath = dotpath.split('.')

    for key in dotpath:
        if (
            not isinstance(d, dict) or \
            key not in d
        ):
            return default

        d = d[key]

    return d

@typecheck
def yaml_config_path_to_dict(
    path: str | Path
) -> dict:

    if isinstance(path, str):
        path = Path(path)

    assert path.is_file(), f'cannot find {str(path)}'

    with open(str(path), 'r') as f:
        maybe_config_dict = yaml.safe_load(f)

    assert exists(maybe_config_dict), f'unable to parse yaml config at {str(path)}'
    assert isinstance(maybe_config_dict, dict), 'yaml config file is not a dictionary'

    return maybe_config_dict

# base pydantic classes for constructing alphafold3 and trainer from config files

class BaseModelWithExtra(BaseModel):
    class Config:
        extra = 'allow'
        use_enum_values = True

class Alphafold3Config(BaseModelWithExtra):
    dim_atom_inputs: int
    dim_template_feats: int
    dim_template_model: int
    atoms_per_window: int
    dim_atom: int
    dim_atompair_inputs: int
    dim_atompair: int
    dim_input_embedder_token: int
    dim_single: int
    dim_pairwise: int
    dim_token: int
    ignore_index: int = -1
    num_dist_bins: int | None
    num_plddt_bins: int
    num_pde_bins: int
    num_pae_bins: int
    sigma_data: int | float
    diffusion_num_augmentations: int
    loss_confidence_weight: int | float
    loss_distogram_weight: int | float
    loss_diffusion_weight: int | float

    @classmethod
    @typecheck
    def from_yaml_file(
        cls,
        path: str | Path,
        dotpath: str | List[str] = []
    ):
        config_dict = yaml_config_path_to_dict(path)
        config_dict = safe_deep_get(config_dict, dotpath)
        assert exists(config_dict), f'config not found at path {".".join(dotpath)}'

        return cls(**config_dict)

    def create_instance(self) -> Alphafold3:
        alphafold3 = Alphafold3(**self.model_dump())
        return alphafold3

    @classmethod
    def create_instance_from_yaml_file(
        cls,
        path: str | Path,
        dotpath: str | List[str] = []
    ) -> Alphafold3:

        af3_config = cls.from_yaml_file(path, dotpath)
        return af3_config.create_instance()

class WeightedPDBSamplerConfig(BaseModelWithExtra):
    chain_mapping_paths: List[FilePath]
    interface_mapping_path: FilePath

    def create_instance(self, batch_size: int):
        return WeightedPDBSampler(**{
            'batch_size': batch_size,
            **self.model_dump()
        })

class DatasetConfig(BaseModelWithExtra):
    dataset_type: Literal['pdb', 'atom'] = 'pdb'
    train_folder: DirectoryPath
    valid_folder: DirectoryPath | None = None
    test_folder: DirectoryPath | None = None
    convert_pdb_to_atom: bool = False
    pdb_to_atom_kwargs: dict = dict()
    train_weighted_sampler: WeightedPDBSamplerConfig | None = None
    kwargs: dict = dict()

class TrainerConfig(BaseModelWithExtra):
    model: Alphafold3Config | None = None
    num_train_steps: int
    batch_size: int
    grad_accum_every: int
    valid_every: int
    ema_decay: float
    lr: float
    clip_grad_norm: int | float
    accelerator: str 
    checkpoint_prefix: str
    checkpoint_every: int
    checkpoint_folder: str
    overwrite_checkpoints: bool
    dataset_config: DatasetConfig | None = None

    @classmethod
    @typecheck
    def from_yaml_file(
        cls,
        path: str | Path,
        dotpath: str | List[str] = []
    ):
        config_dict = yaml_config_path_to_dict(path)
        config_dict = safe_deep_get(config_dict, dotpath)
        assert exists(config_dict), f'config not found at path {".".join(dotpath)}'

        return cls(**config_dict)

    def create_instance(
        self,
        dataset: Dataset | None = None,
        model: Alphafold3 | None = None,
        fabric: Fabric | None = None,
        test_dataset: Dataset | None = None,
        optimizer: Optimizer | None = None,
        scheduler: LRScheduler | None = None,
        valid_dataset: Dataset | None = None,
        map_dataset_input_fn: Callable | None = None,
    ) -> Trainer:

        trainer_kwargs = self.model_dump(
            exclude = {'dataset_config'}
        )

        assert exists(self.model) ^ exists(model), 'either model is available on the trainer config, or passed in when creating the instance, but not both or neither'

        # handle model

        if exists(self.model):
            alphafold3 = self.model.create_instance()
        else:
            alphafold3 = model

        # handle dataset

        if exists(dataset):
            trainer_kwargs.update(dataset = dataset)

        if exists(valid_dataset):
            trainer_kwargs.update(valid_dataset = dataset)

        if exists(test_dataset):
            trainer_kwargs.update(test_dataset = dataset)

        if exists(self.dataset_config):
            dataset_config = self.dataset_config

            dataset_type = dataset_config.dataset_type
            dataset_kwargs = dataset_config.kwargs

            convert_pdb_to_atom = dataset_config.convert_pdb_to_atom
            pdb_to_atom_kwargs = dataset_config.pdb_to_atom_kwargs

            if convert_pdb_to_atom:
                assert dataset_type == 'pdb', 'must be `pdb` dataset_type if `convert_pdb_to_atom` is set to True'

            if dataset_type == 'pdb':
                dataset_klass = PDBDataset
            elif dataset_type == 'atom':
                dataset_klass = AtomDataset
            else:
                raise ValueError(f'unhandled dataset_type {dataset_type}')

            # create dataset for train, valid, and test

            for trainer_kwarg_key, config_key in (('dataset', 'train_folder'), ('valid_dataset', 'valid_folder'), ('test_dataset', 'test_folder')):
                folder = getattr(dataset_config, config_key, None)

                if not exists(folder):
                    continue

                assert trainer_kwarg_key not in trainer_kwargs

                dataset = dataset_klass(folder, **dataset_kwargs)

                if convert_pdb_to_atom:
                    dataset = pdb_dataset_to_atom_inputs(dataset, return_atom_dataset = True, **pdb_to_atom_kwargs)

                trainer_kwargs.update(**{trainer_kwarg_key: dataset})

            # handle weighted pdb sampling

            if exists(dataset_config.train_weighted_sampler):
                sampler = dataset_config.train_weighted_sampler.create_instance(batch_size = self.batch_size)

                trainer_kwargs.update(train_sampler = sampler)

        assert 'dataset' in trainer_kwargs, 'dataset is absent - dataset_type must be specified along with train folders (pdb for now), or the Dataset instance must be passed in'

        # handle rest

        trainer_kwargs.update(dict(
            model = alphafold3,
            fabric = fabric,
            test_dataset = test_dataset,
            optimizer = optimizer,
            scheduler = scheduler,
            valid_dataset = valid_dataset,
            map_dataset_input_fn = map_dataset_input_fn
        ))

        trainer = Trainer(**trainer_kwargs)
        return trainer

    @classmethod
    def create_instance_from_yaml_file(
        cls,
        path: str | Path,
        dotpath: str | List[str] = [],
        **kwargs
    ) -> Trainer:

        trainer_config = cls.from_yaml_file(path, dotpath)
        return trainer_config.create_instance(**kwargs)

# conductor config
# which contains multiple trainer configs for the main and various finetuning stages

class ConductorConfig(BaseModelWithExtra):
    model: Alphafold3Config | None = None
    checkpoint_folder: str
    checkpoint_prefix: str
    training_order: List[str]
    training: Dict[str, TrainerConfig]

    @model_validator(mode = 'after')
    def check_valid_conductor_order(self) -> 'ConductorConfig':
        training_order = set(self.training_order)
        trainer_names = set(self.training.keys())

        if training_order != trainer_names:
            raise ValueError('`training_order` needs to contain all the keys (trainer name) under the `training` field')

        return self

    @classmethod
    @typecheck
    def from_yaml_file(
        cls,
        path: str | Path,
        dotpath: str | List[str] = []
    ):
        config_dict = yaml_config_path_to_dict(path)
        config_dict = safe_deep_get(config_dict, dotpath)
        assert exists(config_dict), f'config not found at path {".".join(dotpath)}'

        return cls(**config_dict)

    def create_instance(
        self,
        trainer_name: str,
        **kwargs
    ) -> Trainer:

        assert trainer_name in self.training, f'{trainer_name} not found among available trainers {tuple(self.training.keys())}'

        model = self.model.create_instance()

        trainer_config = self.training[trainer_name]

        # nest the checkpoint_folder of the trainer within the main checkpoint_folder

        nested_checkpoint_folder = str(Path(self.checkpoint_folder) / Path(trainer_config.checkpoint_folder))

        trainer_config.checkpoint_folder = nested_checkpoint_folder

        # prepend the main training checkpoint_prefix

        nested_checkpoint_prefix = self.checkpoint_prefix + trainer_config.checkpoint_prefix

        trainer_config.checkpoint_prefix = nested_checkpoint_prefix

        # create the Trainer, accounting for root level config

        trainer = trainer_config.create_instance(
            model = model,
            **kwargs
        )

        return trainer

    @classmethod
    def create_instance_from_yaml_file(
        cls,
        path: str | Path,
        dotpath: str | List[str] = [],
        **kwargs
    ) -> Trainer:

        training_config = cls.from_yaml_file(path, dotpath)
        return training_config.create_instance(**kwargs)

# convenience functions

create_alphafold3_from_yaml = Alphafold3Config.create_instance_from_yaml_file
create_trainer_from_yaml = TrainerConfig.create_instance_from_yaml_file
create_trainer_from_conductor_yaml = ConductorConfig.create_instance_from_yaml_file



from random import randrange, random
from dataclasses import asdict

import torch
from torch.utils.data import Dataset
from alphafold3_pytorch import AtomInput

from alphafold3_pytorch.inputs import (
    IS_MOLECULE_TYPES,
    DEFAULT_NUM_MOLECULE_MODS
)

# mock dataset

class MockAtomDataset(Dataset):
    def __init__(
        self,
        data_length,
        max_seq_len = 16,
        atoms_per_window = 4,
        has_molecule_mods = False
    ):
        self.data_length = data_length
        self.max_seq_len = max_seq_len
        self.atoms_per_window = atoms_per_window
        self.has_molecule_mods = has_molecule_mods

    def __len__(self):
        return self.data_length

    def __getitem__(self, idx):
        seq_len = randrange(1, self.max_seq_len)
        atom_seq_len = self.atoms_per_window * seq_len

        atom_inputs = torch.randn(atom_seq_len, 77)
        atompair_inputs = torch.randn(atom_seq_len, atom_seq_len, 5)

        molecule_atom_lens = torch.randint(1, self.atoms_per_window, (seq_len,))
        additional_molecule_feats = torch.randint(0, 2, (seq_len, 5))
        additional_token_feats = torch.randn(seq_len, 2)
        is_molecule_types = torch.randint(0, 2, (seq_len, IS_MOLECULE_TYPES)).bool()

        is_molecule_mod = None
        if self.has_molecule_mods:
            is_molecule_mod = torch.rand((seq_len, DEFAULT_NUM_MOLECULE_MODS)) < 0.05

        molecule_ids = torch.randint(0, 32, (seq_len,))
        token_bonds = torch.randint(0, 2, (seq_len, seq_len)).bool()

        templates = torch.randn(2, seq_len, seq_len, 44)
        template_mask = torch.ones((2,)).bool()

        msa = torch.randn(7, seq_len, 64)

        msa_mask = None
        if random() > 0.5:
            msa_mask = torch.ones((7,)).bool()

        # required for training, but omitted on inference

        atom_pos = torch.randn(atom_seq_len, 3)
        molecule_atom_indices = molecule_atom_lens - 1

        distance_labels = torch.randint(0, 37, (seq_len, seq_len))
        pae_labels = torch.randint(0, 64, (seq_len, seq_len))
        pde_labels = torch.randint(0, 64, (seq_len, seq_len))
        plddt_labels = torch.randint(0, 50, (seq_len,))
        resolved_labels = torch.randint(0, 2, (seq_len,))

        return AtomInput(
            atom_inputs = atom_inputs,
            atompair_inputs = atompair_inputs,
            molecule_ids = molecule_ids,
            token_bonds = token_bonds,
            molecule_atom_lens = molecule_atom_lens,
            additional_molecule_feats = additional_molecule_feats,
            additional_token_feats = additional_token_feats,
            is_molecule_types = is_molecule_types,
            is_molecule_mod = is_molecule_mod,
            templates = templates,
            template_mask = template_mask,
            msa = msa,
            msa_mask = msa_mask,
            atom_pos = atom_pos,
            molecule_atom_indices = molecule_atom_indices,
            distance_labels = distance_labels,
            pae_labels = pae_labels,
            pde_labels = pde_labels,
            plddt_labels = plddt_labels,
            resolved_labels = resolved_labels
        )



"""Amino acid constants used in AlphaFold."""

from typing import Final

import numpy as np

# This mapping is used when we need to store atom data in a format that requires
# fixed atom data size for every residue (e.g. a numpy array).
# From: https://github.com/google-deepmind/alphafold/blob/f251de6613cb478207c732bf9627b1e853c99c2f/alphafold/common/residue_constants.py#L492C1-L497C2
atom_types = [
    "N",
    "CA",
    "C",
    "CB",
    "O",
    "CG",
    "CG1",
    "CG2",
    "OG",
    "OG1",
    "SG",
    "CD",
    "CD1",
    "CD2",
    "ND1",
    "ND2",
    "OD1",
    "OD2",
    "SD",
    "CE",
    "CE1",
    "CE2",
    "CE3",
    "NE",
    "NE1",
    "NE2",
    "OE1",
    "OE2",
    "CH2",
    "NH1",
    "NH2",
    "OH",
    "CZ",
    "CZ2",
    "CZ3",
    "NZ",
    "OXT",
    "_",
    "_",
    "_",
    "_",
    "_",
    "_",
    "_",
    "_",
    "_",
    "_",  # 10 null types.
]
atom_types_set = set(atom_types)
atom_order = {atom_type: i for i, atom_type in enumerate(atom_types)}
atom_type_num = len(atom_types)  # := 37 + 10 null types := 47.
res_rep_atom_index = 1  # The index of the atom used to represent the center of the residue.


# This is the standard residue order when coding AA type as a number.
# Reproduce it by taking 3-letter AA codes and sorting them alphabetically.
restypes = [
    "A",
    "R",
    "N",
    "D",
    "C",
    "Q",
    "E",
    "G",
    "H",
    "I",
    "L",
    "K",
    "M",
    "F",
    "P",
    "S",
    "T",
    "W",
    "Y",
    "V",
]
min_restype_num = 0  # := 0.
restype_order = {restype: min_restype_num + i for i, restype in enumerate(restypes)}
restype_num = min_restype_num + len(restypes)  # := 0 + 20 := 20.


restype_1to3 = {
    "A": "ALA",
    "R": "ARG",
    "N": "ASN",
    "D": "ASP",
    "C": "CYS",
    "Q": "GLN",
    "E": "GLU",
    "G": "GLY",
    "H": "HIS",
    "I": "ILE",
    "L": "LEU",
    "K": "LYS",
    "M": "MET",
    "F": "PHE",
    "P": "PRO",
    "S": "SER",
    "T": "THR",
    "W": "TRP",
    "Y": "TYR",
    "V": "VAL",
    "X": "UNK",
}

BIOMOLECULE_CHAIN: Final[str] = "polypeptide(L)"
POLYMER_CHAIN: Final[str] = "polymer"


# NB: restype_3to1 differs from e.g., Bio.Data.PDBData.protein_letters_3to1
# by being a simple 1-to-1 mapping of 3 letter names to one letter names.
# The latter contains many more, and less common, three letter names as
# keys and maps many of these to the same one letter name
# (including 'X' and 'U' which we don't use here).
restype_3to1 = {v: k for k, v in restype_1to3.items()}

# Define residue metadata for all unknown amino acid residues.
unk_restype = "UNK"
unk_chemtype = "peptide linking"
unk_chemname = "UNKNOWN AMINO ACID RESIDUE"

resnames = [restype_1to3[r] for r in restypes] + [unk_restype]

# This represents the residue chemical type (i.e., `chemtype`) index of amino acid residues.
chemtype_num = 0

# A compact atom encoding with 14 columns for amino acid residues.
# From: https://github.com/google-deepmind/alphafold/blob/f251de6613cb478207c732bf9627b1e853c99c2f/alphafold/common/residue_constants.py#L505
# Also matches: https://files.rcsb.org/ligands/view/ALA.cif - https://files.rcsb.org/ligands/view/VAL.cif
# pylint: disable=line-too-long
# pylint: disable=bad-whitespace
restype_name_to_compact_atom_names = {
    "ALA": ["N", "CA", "C", "O", "CB", "", "", "", "", "", "", "", "", ""],
    "ARG": ["N", "CA", "C", "O", "CB", "CG", "CD", "NE", "CZ", "NH1", "NH2", "", "", ""],
    "ASN": ["N", "CA", "C", "O", "CB", "CG", "OD1", "ND2", "", "", "", "", "", ""],
    "ASP": ["N", "CA", "C", "O", "CB", "CG", "OD1", "OD2", "", "", "", "", "", ""],
    "CYS": ["N", "CA", "C", "O", "CB", "SG", "", "", "", "", "", "", "", ""],
    "GLN": ["N", "CA", "C", "O", "CB", "CG", "CD", "OE1", "NE2", "", "", "", "", ""],
    "GLU": ["N", "CA", "C", "O", "CB", "CG", "CD", "OE1", "OE2", "", "", "", "", ""],
    "GLY": ["N", "CA", "C", "O", "", "", "", "", "", "", "", "", "", ""],
    "HIS": ["N", "CA", "C", "O", "CB", "CG", "ND1", "CD2", "CE1", "NE2", "", "", "", ""],
    "ILE": ["N", "CA", "C", "O", "CB", "CG1", "CG2", "CD1", "", "", "", "", "", ""],
    "LEU": ["N", "CA", "C", "O", "CB", "CG", "CD1", "CD2", "", "", "", "", "", ""],
    "LYS": ["N", "CA", "C", "O", "CB", "CG", "CD", "CE", "NZ", "", "", "", "", ""],
    "MET": ["N", "CA", "C", "O", "CB", "CG", "SD", "CE", "", "", "", "", "", ""],
    "PHE": ["N", "CA", "C", "O", "CB", "CG", "CD1", "CD2", "CE1", "CE2", "CZ", "", "", ""],
    "PRO": ["N", "CA", "C", "O", "CB", "CG", "CD", "", "", "", "", "", "", ""],
    "SER": ["N", "CA", "C", "O", "CB", "OG", "", "", "", "", "", "", "", ""],
    "THR": ["N", "CA", "C", "O", "CB", "OG1", "CG2", "", "", "", "", "", "", ""],
    "TRP": [
        "N",
        "CA",
        "C",
        "O",
        "CB",
        "CG",
        "CD1",
        "CD2",
        "NE1",
        "CE2",
        "CE3",
        "CZ2",
        "CZ3",
        "CH2",
    ],
    "TYR": ["N", "CA", "C", "O", "CB", "CG", "CD1", "CD2", "CE1", "CE2", "CZ", "OH", "", ""],
    "VAL": ["N", "CA", "C", "O", "CB", "CG1", "CG2", "", "", "", "", "", "", ""],
    "UNK": ["N", "CA", "C", "O", "CB", "", "", "", "", "", "", "", "", ""],
}

restype_atom47_to_compact_atom = np.zeros([21, 47], dtype=int)


def _make_constants():
    """Fill the array(s) above."""
    for restype, restype_letter in enumerate(restypes):
        resname = restype_1to3[restype_letter]
        for compact_atomidx, atomname in enumerate(restype_name_to_compact_atom_names[resname]):
            if not atomname:
                continue
            atomtype = atom_order[atomname]
            restype_atom47_to_compact_atom[restype, atomtype] = compact_atomidx


_make_constants()
